// Emacs style mode select   -*- C++ -*-
// -----------------------------------------------------------------------------
// ########   ###### #####   #####  ######   ######  ######
// ##     ##  ##     ##  ## ##  ## ##    ## ##    ## ##   ##
// ##     ##  ##     ##   ###   ## ##    ## ##    ## ##    ##
// ########   ####   ##    #    ## ##    ## ##    ## ##    ##
// ##    ##   ##     ##         ## ##    ## ##    ## ##    ##
// ##     ##  ##     ##         ## ##    ## ##    ## ##   ##
// ##      ## ###### ##         ##  ######   ######  ######
//                      http://remood.sourceforge.net/
// -----------------------------------------------------------------------------
// Project Leader:    GhostlyDeath           (ghostlydeath@gmail.com)
// Project Co-Leader: RedZTag                (jostol27@gmail.com)
// Members:           Demyx                  (demyx@endgameftw.com)
// -----------------------------------------------------------------------------
// Copyright (C) 2008-2009 The ReMooD Team.
// -----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// -----------------------------------------------------------------------------
// DESCRIPTION: Master Server

#include "global.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>

// Network globals
int NetSocket = -1;
struct sockaddr_in SocketControl;

/* NET_BindPort() -- Binds a UDP Port */
NetworkError_t NET_BindPort(UInt16 Port)
{
	NetSocket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	
	if (NetSocket == -1)
		return NE_SOCKETFAILURE;
		
	SocketControl.sin_addr.s_addr = INADDR_ANY;
	SocketControl.sin_port = htons(Port);
		
	if (bind(NetSocket, (struct sockaddr*)&SocketControl, sizeof(struct sockaddr)) == -1)
	{
		close(NetSocket);
		return NE_BINDTOPORTERROR;
	}
	
	return NE_SUCCESS;
}

/* NET_ClosePort() -- Closes the UDP Port */
NetworkError_t NET_ClosePort(void)
{
	if (NetSocket == -1)
		return NE_SOCKETFAILURE;
	
	close(NetSocket);
	NetSocket = -1;
	
	return NE_SUCCESS;
}

/* NET_RecvData() -- Recieves UDP Data */
NetworkError_t NET_RecvData(Host_t* From, void* BufferPtr, size_t BufferSize, size_t* RecvSize)
{
	struct sockaddr_in FromSock;
	UInt32 FromLen = sizeof(struct sockaddr*);
	ssize_t len = 0;
	
	if (BufferSize == 0)
		return NE_BUFFERTOOSMALL;
	else if (!BufferPtr)
		return NE_INVALIDPARAMETER;
	
	if ((len = recvfrom(NetSocket, BufferPtr, BufferSize, MSG_DONTWAIT, (struct sockaddr*)&FromSock, &FromLen)) > 0)
	{
		if (From)
		{
			From->ip.dw = ntohl(FromSock.sin_addr.s_addr);
			From->port = ntohs(FromSock.sin_port);
		}
		
		if (RecvSize)
			*RecvSize = len;
		
		return NE_SUCCESS;
	}
	else
		return NE_BUFFEREMPTY;
}

/* NET_SendData() -- Sends UDP Data */
NetworkError_t NET_SendData(Host_t* To, void* BufferPtr, size_t BufferSize, size_t* SentSize)
{
	struct sockaddr_in ToSock;
	ssize_t len;
	
	if (!BufferSize || !BufferPtr)
		return NE_INVALIDPARAMETER;
	
	if (To)
	{
		ToSock.sin_addr.s_addr = htonl(To->ip.dw);
		ToSock.sin_port = htons(To->port);
	}
	
	if ((len = sendto(NetSocket, BufferPtr, BufferSize, 0, (struct sockaddr*)&ToSock, sizeof(struct sockaddr*))) > 0)
	{
		if (SentSize)
			*SentSize = len;
		
		return NE_SUCCESS;
	}
	else
		return NE_BUFFEREMPTY;
}

/* M_MasterMessage() -- Print a message from the master */
MasterError_t M_MasterMessage(char* Message, ...)
{
	va_list argptr;
	char Text[512];
	char Time[32];
	time_t TheTime = time(NULL);
	struct tm* LocalTime = localtime(&TheTime);
	
	va_start(argptr, Message);
	vsnprintf(Text, sizeof(Text) / sizeof(char), Message, argptr);
	va_end(argptr);
	
	strftime(Time, sizeof(Time) / sizeof(char), "%Y/%m/%d %H:%M:%S %Z", LocalTime);
	
	printf("[%s] %s\n", Time, Text);
	
	return ME_SUCCESS;
}

/* MasterInterface() -- Master Server Control */
void* MasterInterface(void* arg)
{
	clock_t intime = 0;
	clock_t outtime = 0;
	UInt32 sleeptime = 0;
	double grans = 1;
	
	// Determine granularity (for milliseconds time)
	sleeptime = CLOCKS_PER_SEC;
	
	while ((sleeptime * grans) > 1000)
		grans /= 10;
	
	while ((sleeptime * grans) < 1000)
		grans *= 10;
	
	if (M_StartMaster())
		return NULL;
	
	// The Loop iterates every few seconds
	for (; M_Status() != ME_SHUTDOWN && M_Status() != ME_OFFLINE;)
	{
		intime = clock();
		
		M_Iterate();
		
		outtime = clock();
		
		// Sleep only if we can afford to
		sleeptime = 2500 - ((outtime - intime) / (CLOCKS_PER_SEC * grans));
		if ((sleeptime > 0) && (sleeptime <= 7500))	// in case of clock skew
				usleep(sleeptime * 1000);
	}
	
	if (M_StopMaster())
		return NULL;
		
	return NULL;
}

#define COMMANDBUFFERSIZE 512

/* main() -- Main Program */
int main(int argc, char** argv)
{
	pthread_t MasterThread;
	char* Buffer = NULL;
	char* TokenBuffer = NULL;
	char* c = NULL;
	char* d = NULL;
	Int8 mode = 0;
	size_t i;
	
	printf("########   ###### #####   #####  ######   ######  ######\n");
	printf("##     ##  ##     ##  ## ##  ## ##    ## ##    ## ##   ##\n");
	printf("##     ##  ##     ##   ###   ## ##    ## ##    ## ##    ##\n");
	printf("########   ####   ##    #    ## ##    ## ##    ## ##    ##\n");
	printf("##    ##   ##     ##         ## ##    ## ##    ## ##    ##\n");
	printf("##     ##  ##     ##         ## ##    ## ##    ## ##   ##\n");
	printf("##      ## ###### ##         ##  ######   ######  ######\n");
	printf("\n");
	printf("ReMooD Master Server v1.0\n");
	printf("Copyright (C) 2008 GhostlyDeath (ghostlydeath@gmail.com).\n");
	printf("\n");
	printf("This program is free software; you can redistribute it and/or\n");
	printf("modify it under the terms of the GNU General Public License\n");
	printf("as published by the Free Software Foundation; either version 2\n");
	printf("of the License, or (at your option) any later version.\n");
	printf("\n");
	printf("This program is distributed in the hope that it will be useful,\n");
	printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	printf("GNU General Public License for more details.\n");
	printf("\n");
	
	// Start master thread
	if (pthread_create(&MasterThread, NULL, MasterInterface, NULL))
		return EXIT_FAILURE;
	
	// Prepare buffers
	Buffer = malloc(COMMANDBUFFERSIZE);
	TokenBuffer = malloc(COMMANDBUFFERSIZE);
	if (!Buffer || !TokenBuffer)
	{
		if (Buffer)
			free(Buffer);
		if (TokenBuffer)
			free(TokenBuffer);
		
		M_StopMaster();
		return EXIT_FAILURE;
	}
	
	memset(Buffer, 0, COMMANDBUFFERSIZE);
	
	// Command loop
	for (;;)
	{
		// Get STDIN
		if (fgets(Buffer, COMMANDBUFFERSIZE, stdin) != EOF)
		{
			// Tokenize, kind of
			memset(TokenBuffer, 0, COMMANDBUFFERSIZE);
			
			c = Buffer;
			d = TokenBuffer;
			
			while (*c)
			{
				if (*c != ' ' && *c != '\r' && *c != '\n')
					*d = *c;
				
				c++;
				d++;
			}
			
			TokenBuffer[COMMANDBUFFERSIZE - 1] = 0;
			TokenBuffer[COMMANDBUFFERSIZE - 2] = 0;
			
			if (strlen(TokenBuffer) == 0)
				continue;
			
			printf("~> %s\n", TokenBuffer);
			
			/* Check commands */
				// EXIT or QUIT
			if ((strcasecmp(TokenBuffer, "exit") == 0) || (strcasecmp(TokenBuffer, "quit") == 0))
			{
				printf("Shutting down, please wait...\n");
				M_StopMaster();
				break;
			}
			else
			{
				printf("Invalid command or variable!\n");
			}
		}
	}
	
	while (M_Status() != ME_OFFLINE)
		usleep(100000);	// sleep for 100ms then try again so we don't fry the CPU
	
	return EXIT_SUCCESS;
}

